try:
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
   pass

from flask import (
    abort,
    Blueprint,
    jsonify,
    request,
    current_app,
)
from flask_jwt_extended import (
    jwt_required,
    get_jwt_identity,
)
from dtool_lookup_server import AuthenticationError

from .config import Config
from .utils import (
    config_to_dict,
    query_datasets_by_user,
    aggregate_datasets_by_user,
)


mongo_bp = Blueprint("mongo", __name__, url_prefix="/mongo")


@mongo_bp.route("/config", methods=["GET"])
@jwt_required()
def plugin_config():
    """Return the JSON-serialized plugin configuration."""
    username = get_jwt_identity()
    try:
        config = config_to_dict(username)
    except AuthenticationError:
        abort(401)
    return jsonify(config)


@mongo_bp.route("/query", methods=["POST"])
@jwt_required()
def query_datasets():
    """Query datasets a user has access to."""
    if not Config.ALLOW_DIRECT_QUERY:
        abort(404)
    username = get_jwt_identity()
    query = request.get_json()
    current_app.logger.debug("Received query request '{}' from user '{}'.".format(query, username))
    try:
        datasets = query_datasets_by_user(username, query)
    except AuthenticationError:
        abort(401)
    return jsonify(datasets)


@mongo_bp.route("/aggregate", methods=["POST"])
@jwt_required()
def aggregate_datasets():
    """Aggregate the datasets a user has access to."""
    if not Config.ALLOW_DIRECT_AGGREGATION:
        abort(404)
    username = get_jwt_identity()
    query = request.get_json()
    current_app.logger.debug("Received aggregate request '{}' from user '{}'.".format(query, username))
    try:
        datasets = aggregate_datasets_by_user(username, query)
    except AuthenticationError:
        abort(401)
    return jsonify(datasets)



"""Test the /mongo/config blueprint route."""

import json
import dtool_lookup_server_direct_mongo_plugin

from . import tmp_app_with_users  # NOQA

from . import snowwhite_token

def test_config_info_route(tmp_app_with_users):  # NOQA

    headers = dict(Authorization="Bearer " + snowwhite_token)
    r = tmp_app_with_users.get(
        "/mongo/config",
        headers=headers,
    )
    assert r.status_code == 200

    expected_content = {
        'allow_direct_query': True,
        'allow_direct_aggregation': False,
        'version': dtool_lookup_server_direct_mongo_plugin.__version__}

    response = json.loads(r.data.decode("utf-8"))
    assert response == expected_content



import os
from . import __version__


class Config(object):
    # Above all enters the flask app, below does not.
    # Would those settings belong here or should they be somewhere else, i.e.
    # a specific dtool lookup server configuration file?
    # This option allows a client to submit direct mongo-syntaxed queries
    # to the underlying mongo database. Externally managed privileges will
    # be enforced as usual by embedding such queries in accompanying logical
    # 'and' clauses, see utils._preprocess_privileges() and
    #  utils._dict_to_mongo_query().
    ALLOW_DIRECT_QUERY = os.environ.get('DTOOL_LOOKUP_SERVER_ALLOW_DIRECT_QUERY',
                                        'True').lower() in ['true', '1', 'y', 'yes', 'on']
    # This option allows a client to submit direct mongo-syntaxed aggregations
    # to the underlying mongo database. As above, externally managed privileges
    # will still apply to the initial '$match' stage of the aggregation
    # pipeline (see utils._dict_to_mongo_aggregation()), but can be easiliy
    # circumvented in subsequent aggregation stages. Further notice that
    # aggregation stages allow write access to the database, thus this option
    # should only be enabled if some privileges are configured a the MongoDB
    # level as well.
    ALLOW_DIRECT_AGGREGATION = os.environ.get('DTOOL_LOOKUP_SERVER_ALLOW_DIRECT_AGGREGATION',
                                              'False').lower() in ['true', '1', 'y', 'yes', 'on']

    @classmethod
    def to_dict(cls):
        """Convert plugin configuration into dict."""
        exclusions = []  # config keys to exclude
        d = {'version': __version__}
        for k, v in cls.__dict__.items():
            # select only capitalized fields
            if k.upper() == k and k not in exclusions:
                d[k.lower()] = v
        return d



from setuptools import setup
from setuptools_scm import get_version
version = get_version(root='.', relative_to=__file__)


def local_scheme(version):
    """Skip the local version (eg. +xyz of 0.6.1.dev4+gdf99fe2)
    to be able to upload to Test PyPI"""
    return ""


url = "https://github.com/IMTEK-Simulation/dtool-lookup-server-direct-mongo-plugin"
readme = open('README.rst').read()

setup(
    name="dtool-lookup-server-direct-mongo-plugin",
    packages=["dtool_lookup_server_direct_mongo_plugin"],
    description="""This plugin allows to submit mongo queries and aggregation
                   pipelines directly to the lookup server's underlying MongoDB.""",
    long_description=readme,
    author="Johannes Hörmann",
    author_email="johannes.hoermann@imtek.uni-freiburg.de",
    use_scm_version={"local_scheme": local_scheme},
    url=url,
    entry_points={
        'dtool_lookup_server.blueprints': [
            'dtool_lookup_server_direct_mongo_plugin=dtool_lookup_server_direct_mongo_plugin:mongo_bp',
        ],
    },
    setup_requires=['setuptools_scm'],
    install_requires=[
        "dtool-lookup-server>=0.17.2",
        "dtoolcore>=3.17.0",
    ],
    download_url="{}/tarball/{}".format(url, version),
    license="MIT"
)



"""Utility functions."""
# TODO: expose protected members _preprocess_privileges and _dict_to_mongo_query
#       of dtool_lookup_server.utils publically for plugin API.

import logging

from flask import current_app
from dtool_lookup_server import mongo, MONGO_COLLECTION

import dtoolcore.utils
import dtool_lookup_server.utils

from .config import Config

logger = logging.getLogger(__name__)


def config_to_dict(username):
    # TODO: check on privileges
    return Config.to_dict()


def _dict_to_mongo_query(query_dict):
    """Construct mongo query as usual, but allow embedding a raw mongo query.

    Treat query_dict as in dtool_lookup_server.utils._dict_to_mongo_query, but
    additionally embed raw mongo query if key 'query' exists."""

    if "query" in query_dict and isinstance(query_dict["query"], dict):
        raw_mongo = query_dict["query"]
        del query_dict["query"]
    else:
        raw_mongo = {}

    # NOTE: This requires knowledge about the structure of what the original
    # dtool_lookup_server.utils._dict_to_mongo_query returns:
    mongo_query = dtool_lookup_server.utils._dict_to_mongo_query(query_dict)

    if len(raw_mongo) > 0 and len(mongo_query) == 0:
        mongo_query = raw_mongo
    elif len(raw_mongo) > 0 and len(mongo_query) == 1 and "$and" in mongo_query:
        mongo_query["$and"].append(raw_mongo)
    elif len(raw_mongo) > 0:
        mongo_query = {"$and": [mongo_query, raw_mongo]}

    logger.debug("Constructed mongo query: {}".format(mongo_query))
    return mongo_query


def _dict_to_mongo_aggregation(query_dict):
    """Construct mongo query as usual and prepend to aggregation pipeline."""
    if "aggregation" in query_dict and isinstance(query_dict["aggregation"], list):
        aggregation_tail = query_dict["aggregation"]
        del query_dict["aggregation"]
    else:
        aggregation_tail = []

    # unset any _id field, as type ObjectId usually not serializable
    aggregation_tail.append({'$unset': '_id'})

    match_stage = _dict_to_mongo_query(query_dict)
    if len(match_stage) > 0:
        aggregation_head = [{'$match': match_stage}]
    else:
        aggregation_head = []

    aggregation = [*aggregation_head, *aggregation_tail]
    current_app.logger.debug("Constructed mongo aggregation: {}".format(aggregation))
    return aggregation


def query_datasets_by_user(username, query):
    """Query the datasets the user has access to. Allow raw mongo 'query'.

    See dtool_lookup_server.utils.search_datasets_by_user docstring.

    :param username: username
    :param query: dictionary specifying query
    :returns: List of dicts if user is valid and has access to datasets.
              Empty list if user is valid but has not got access to any
              datasets.
    :raises: AuthenticationError if user is invalid.
    """

    query = dtool_lookup_server.utils._preprocess_privileges(username, query)
    # If there are no base URIs at this point it means that the user is not
    # allowed to search for anything.
    if len(query["base_uris"]) == 0:
        return []

    datasets = []
    mongo_query = _dict_to_mongo_query(query)
    cx = mongo.db[MONGO_COLLECTION].find(
        mongo_query,
        {
            "_id": False,
            "readme": False,
            "manifest": False,
            "annotations": False,
        }
    )
    for ds in cx:

        # Convert datetime object to float timestamp.
        for key in ("created_at", "frozen_at"):
            if key in ds:
                datetime_obj = ds[key]
                ds[key] = dtoolcore.utils.timestamp(datetime_obj)

        datasets.append(ds)
    return datasets


def aggregate_datasets_by_user(username, query):
    """Aggregate the datasets the user has access to.
    Valid keys for the query are: creator_usernames, base_uris, free_text,
    aggregation. If the query dictionary is empty, all datasets that a user has
    access to are returned.
    :param username: username
    :param query: dictionary specifying query
    :returns: List of dicts if user is valid and has access to datasets.
              Empty list if user is valid but has not got access to any
              datasets.
    :raises: AuthenticationError if user is invalid.
    """
    if not Config.ALLOW_DIRECT_AGGREGATION:
        current_app.logger.warning(
            "Received aggregate request '{}' from user '{}', but direct "
            "aggregations are disabled.".format(query, username))
        return []  # silently reject request

    # TODO: make applying privileges configurable
    query = dtool_lookup_server.utils._preprocess_privileges(username, query)

    # If there are no base URIs at this point it means that the user has not
    # got privileges to search for anything.
    # TODO: reject on blueprint level (?)
    if len(query["base_uris"]) == 0:
        return []
    datasets = []

    mongo_aggregation = _dict_to_mongo_aggregation(query)
    cx = mongo.db[MONGO_COLLECTION].aggregate(mongo_aggregation)
    # Opposed to search_datasets_by_user, here it is the aggregator's
    # responsibility to project out desired fields and remove non-serializable
    # content. The only modification always applied is removing any '_id' field.
    for ds in cx:
        # Convert datetime object to float timestamp.
        for key in ("created_at", "frozen_at"):
            if key in ds:
                datetime_obj = ds[key]
                ds[key] = dtoolcore.utils.timestamp(datetime_obj)

        datasets.append(ds)

    return datasets
